/*
************************************************************
* sndctrl.cc                                               *
* extended functions to control soundcards                 *
* last change 2000-10-16                                   *
* Robin Gruber <archi@robinsnest.de>                       *
* for details have a look at the README file               *
************************************************************
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>                                           
#include <termios.h>
#include <math.h>                                           
#include "sndctrl.h"
#include "debug.h"
#include "../lib/micpos.h"                                           
#include "../lib/soundlow.h"                                           

#define DEBUG(x) 


// these are the controls for the simulation (theta and phi of the virtual sound-source)
Control::Control(QWidget * parent=NULL, const char * name=NULL)
    : QFrame (parent, name)
{
    // generate a scrollbar for phi...
    phiBar=new QScrollBar (-180,180,1,10,0,QScrollBar::Horizontal,this,"phiBar");
    phiBar->setGeometry(5,25,200,15);
    // ...and connect it to the phi-handler
    connect(phiBar,SIGNAL(valueChanged(int)),this,SLOT(phiBarChanged(int)));
    // generate a scrollbar for theta...
    thetaBar=new QScrollBar (-90,90,1,10,0,QScrollBar::Horizontal,this,"thetaBar");
    thetaBar->setGeometry(5,65,200,15);
    // ...and connect it to the theta-handler
    connect(thetaBar,SIGNAL(valueChanged(int)),this,SLOT(thetaBarChanged(int)));
    // initialize angles with zero
    theta=0;
    phi=0;
    // set position fixed size for the control-frame
    setFixedSize(210,90);
    setGeometry(400,85,width(),height());
    show();
}


Control::~Control()
{
    delete phiBar;
    delete thetaBar;
}

// the scrollbar for phi is changed
void Control::phiBarChanged(int p)
{
    // update angle
    phi=p;
    // and repaint the frame, so that the angle is diplayed correctly
    repaint();
}

// the scrollbar for theta is changed
void Control::thetaBarChanged(int t)
{
    // update angle
    theta=t;
    // and repaint the frame, so that the angle is diplayed correctly
    repaint();
}

// redraw the simulation-control frame
void Control::paintEvent(QPaintEvent *)
{
    // the bars are being repainted automatically,
    // so we have to repaint only the description
    QPainter p;
    p.begin(this);
    char s[20];
    sprintf(s,"Phi:   %4d",phi);
    p.drawText(15,20,s);
    sprintf(s,"Theta: %4d",theta);
    p.drawText(15,60,s);
    p.end();
}

// count cares, that only one Sound object can be created
int Sound::count=0;

Sound::Sound(int c, int r, int m, char * inputfilename, QObject * parent=NULL, char * name=NULL) : QObject(parent, name )
{
    int i;
    double f;
    char s[2000];

    // check if more than one sound object generated
    count++;
    DEBUG(fprintf(stderr,"Sound: Constructor called\n"));
    if (count>1) {
	fprintf(stderr,"Sound: error: Two sound objects generated!\n");
	exit(1);
    }

    // initialize variables
    channels=c;
    rate=r;
    simmode=m;
    samplingActive=false;
    cards=(channels+1)/2;
    DEBUG(fprintf(stderr,"Sound: Using %d cards!\n",cards));
    
    // allocate the sound-data-fifo
    buffer=(signed short*)malloc(BUFFERSIZE*READSIZE*2*cards*sizeof(short));
    if (buffer==NULL) {
        fprintf(stderr,"Sound: cannot allocate enough memory for buffer!\n");
        exit(1);
    }

    // if file-io mode is desired, open the sound-files!
    if (simmode==MODE_FILEIO) {
        iofile=(int*)malloc(channels*sizeof(int));
        for (i=0; i<channels; i++) {
            sprintf(s,"%s%02d.raw",inputfilename,i);
            if ((iofile[i]=open(s,O_RDONLY, 0644))==-1) {
                perror("Sound: Error opening soundfiles:");
                exit(1);
            }
        }
        DEBUG(fprintf(stderr,"Sound: soundfiles opened!\n"));
    }
    // if simulation-mode is desired, setup controls and variables
    if (simmode==MODE_SIMULATION) {
        control = new Control();
        control->setCaption("Simulation: Angles");
        connect(control->phiBar,SIGNAL(valueChanged(int)),this,SLOT(setSimPhi(int)));
        connect(control->thetaBar,SIGNAL(valueChanged(int)),this,SLOT(setSimTheta(int)));
        simPos=0;
        simPhi=0.0;
        simTheta=0.0;
        // preset an array with calculated sound
        for (i=0; i<SIMLENGTH; i++) {
            f=double(i)/double(SIMLENGTH)*2.0*M_PI;
            wave[i]=(signed short)((sin(f*2)+1.5)*(sin(f*200.0)*3000.0+sin(f*330.0)*5000
                                                   +cos(f*133.0)*3000+cos(f*550)*1000
                                                   -sin(f*4946)*2500-cos(f*2143)*2000));
            if (wave[i]>32000) {wave[i]=32000;}
            if (wave[i]<-32000) {wave[i]=-32000;}
          
            // wave[i]=rand()/(RAND_MAX/32768)-16384;
            // wave[i]=int(sin(f*800.0)*8000);
            // wave[i]=0;
            //wave[i]=(signed short)((sin(f*500)+sin(f*1500))*16000);
        }
    }

    // if normal soundcard-operation is desired, just initialize soundcard!
    if (simmode==MODE_NORMAL) {
        soundcard=new SoundCard(cards,rate,READSIZE);
    }
}


Sound::~Sound()
{
    DEBUG(fprintf(stderr,"Sound: Destructor called!\n"));
}


void Sound::startSampling()
{
    int delaytime;
    samplingActive=true;
    bufferPos=0;
    // start the soundcards if nessesairy
    if (simmode==MODE_NORMAL) {soundcard->go();}
    // we use polling, so install a timer, that triggers when data
    // is ready
    delaytime=READSIZE*1000/rate;
    delaytime=delaytime-PRETRIGGER;
    DEBUG(fprintf(stderr,"Sound: Delaytime is %d ms!\n",delaytime));
    startTimer(delaytime);
}


signed short Sound::getSimValue(int nr,int l)
{
    double d;
    // Turn the simulated mircophone virtually to calculate the delay
    d=(cos(simPhi)*(cos(simTheta)*micpos[nr].x+sin(simTheta)*micpos[nr].z)
       -sin(simPhi)*micpos[nr].y)/double(SPEEDOFSOUND)*double(rate);
    return wave[(simPos+int(rint(d))+l+SIMLENGTH)%SIMLENGTH];
}


// slots for the simulation-control bars
void Sound::setSimTheta(int angle)
{
    simTheta=double(angle)/90.0*M_PI;
    printf("Theta:%f\n",simTheta);
}

void Sound::setSimPhi(int angle)
{
    simPhi=double(angle)/180.0*M_PI;
    printf("Phi:%f\n",simPhi);
}


// timer-event: sound data should be available now!
void Sound::timerEvent(QTimerEvent *)
{
    int l,i;
    signed short dummy[READSIZE];
    DEBUG(fprintf(stderr,"Sound: Timer Signal received!\n"));
    DEBUG(fprintf(stderr,"Sound: Bufferpos is %d\n",bufferPos));

    // normal mode: read data from soundcards
    if (simmode==MODE_NORMAL) {	
        soundcard->read(&buffer[bufferPos*cards*READSIZE*2]);
        DEBUG(fprintf(stderr,"Sound: Data read\n"));
    }

    //simulation mode: simulate data
    if (simmode==MODE_SIMULATION) {
        for (i=0; i<cards; i++) {
            for (l=0; l<READSIZE; l++) {
                buffer[((bufferPos*cards+i)*READSIZE+l)*2]=getSimValue(i*2,l);
                buffer[((bufferPos*cards+i)*READSIZE+l)*2+1]=getSimValue(i*2+1,l);
            }
        }
    }
    // fileio mode: read data from files
    if (simmode==MODE_FILEIO) {
        for (i=0; i<cards; i++) {
            if (read(iofile[i*2],dummy,READSIZE*2)!=READSIZE*2) {
                fprintf(stderr,"Sound: inputfile empty!\n");
                exit(0);
            }
            for (l=0; l<READSIZE; l++) {
                buffer[((bufferPos*cards+i)*READSIZE+l)*2]=dummy[l];
            }
            if (read(iofile[i*2+1],dummy,READSIZE*2)!=READSIZE*2) {
                fprintf(stderr,"Sound: inputfile empty!\n");
                exit(0);
            }
            for (l=0; l<READSIZE; l++) {
                buffer[((bufferPos*cards+i)*READSIZE+l)*2+1]=dummy[l];
            }
        }
    }    
    simPos+=READSIZE;
    simPos%=SIMLENGTH;
    bufferPos=(bufferPos+1)%BUFFERSIZE;
    // emit a signal to every plugin that sound-data could be available
    emit newDataReady();
    DEBUG(fprintf(stderr,"Sound:End of timer-routine!\n"));
}


inline int Sound::getBufferPos()
{
    return bufferPos*READSIZE;
}


inline signed short Sound::getBuffer(int channel, int pos)
{
    // fixme:: documentaion of the tail below!!!
    if (channel>=channels) {return 0;}
    return buffer[((pos/READSIZE)%BUFFERSIZE)*(cards<<1)*READSIZE
		 +(channel>>1)*(READSIZE<<1)+
		 ((pos%READSIZE)<<1)+(channel&1)];
}

int Sound::getChannels()
{
    return channels;
}

int Sound::getSampleRate()
{
    return rate;;
}

// each plugin generates a new sound-receiver object
SoundReceiver::SoundReceiver(Sound * app, int c)
{
    int i;
    soundApp=app;
    DEBUG(fprintf(stderr,"SoundReceiver: Constructor called\n"));
    channels=c;
    delays=(int *)malloc(channels*sizeof(int));
    data=(short signed**)malloc(channels*sizeof(short*));
    for (i=0; i<channels; i++) {
	delays[i]=0;
	data[i]=NULL;
    }
    frameSize=0;
    stepSize=0;
    readPos=0;
    connect(soundApp,SIGNAL(newDataReady()),this,SLOT(dataReady()));
}

SoundReceiver::~SoundReceiver()
{
    int i;
    free(delays);
    for (i=0; i<channels; i++) {
	if (data[i]!=NULL) {free(data[i]);}
    }
    free(*data);
}

// this object is called every time data is read from the soundcard
void SoundReceiver::dataReady()
{
    int i,k,dataPos;
    DEBUG(fprintf(stderr,"SoundReceiver: dataReady() Signal received!\n"));
    if (frameSize<=0) {return;}
    if (stepSize<=0) {return;}
    while (1) {
	dataPos=soundApp->getBufferPos();
	if (dataPos<readPos) {dataPos+=BUFFERSIZE*READSIZE;}
	DEBUG(fprintf(stderr,"SoundReceiver: readPos: %d, dataPos: %d\n",readPos,dataPos));
	for (i=0; i<channels; i++) {
           DEBUG(fprintf(stderr,"SoundReceiver: testing channel %d\n",i));
           if (readPos+frameSize+delays[i]>=dataPos){return;}
           if (readPos+stepSize>=dataPos){return;}
	}
        DEBUG(fprintf(stderr,"SoundReceiver: filling local buffer!\n"));
        // get the samples and fill them in the local buffer
	for (i=0; i<channels; i++) {
	    for (k=0; k<frameSize; k++) {
	      data[i][k]=soundApp->getBuffer(i,k+delays[i]+readPos);
	    }
	}
        // emit signal to the plugin, data is ready in the desired size
	DEBUG(fprintf(stderr,"SoundReceiver: Emitting Signal!\n"));
	emit newDataReady(channels,frameSize, data);
	readPos=(readPos+stepSize)%(BUFFERSIZE*READSIZE);
    }
}

// each channel can have a desired delay
void SoundReceiver::setDelay(int delayNr, int delay)
{
    if (delayNr<0) {
	fprintf(stderr,"Soundreceiver: error: Negative delay requested!\n");
	exit(1);
    }
    if (delayNr<channels) {
	delays[delayNr]=delay;
	DEBUG(fprintf(stderr,"Delay for channel %d is %d\n",delayNr,delays[delayNr]));
    }
    else {
	fprintf(stderr,"Soundreceiver: warning: Delay for nonexistant channel set!\n");
    }
}

// the data-frames can have indiviual sizes
void SoundReceiver::setFrameSize(int size)
{
    int i;
    frameSize=size;
    DEBUG(fprintf(stderr,"Soundreceiver: frameSize set to %d!\n",frameSize));
    for (i=0; i<channels; i++) {
	data[i]=(signed short*)realloc(data[i],frameSize*sizeof(short));
    }
}

// the data-frames can have individual overlapping/distances
void SoundReceiver::setStepSize(int size)
{
    stepSize=size;
    DEBUG(fprintf(stderr,"Soundreceiver: stepSize set to %d!\n",frameSize));
}
